/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <string.h>
#include <limits.h>
#include <mad.h>
#include <unistd.h>
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>


#define BITS 16
#define INBUF_SIZE 8192
#define OUTBUF_SIZE 8192
#define MAD_TO_SHORT(f) (f) >= MAD_F_ONE?  SHRT_MAX :\
(f) <= -MAD_F_ONE? -SHRT_MAX : (signed short) ((f) >> (MAD_F_FRACBITS - 15))

static unsigned char inbuf[INBUF_SIZE + MAD_BUFFER_GUARD], outbuf[OUTBUF_SIZE];
static struct mad_stream stream;

/* derived from oggdec.c of the vorbis-tools-1.0.1 package */
static int write_wav_header(int channels, int samplerate)
{
#define WRITE_U32(buf, x) (buf)[0]= (unsigned char)((x)&0xff);\
	(buf)[1] = (unsigned char)(((x)>>8)&0xff);\
	(buf)[2] = (unsigned char)(((x)>>16)&0xff);\
	(buf)[3] = (unsigned char)(((x)>>24)&0xff);
#define WRITE_U16(buf, x) (buf)[0] = (unsigned char)((x)&0xff);

	static unsigned char headbuf[44];
	unsigned int size = 0x7fffffff;
	int bytespersec = channels * samplerate * BITS / 8;
	int align = channels * BITS / 8;

	memcpy(headbuf, "RIFF", 4);
	WRITE_U32(headbuf + 4, size - 8);
	memcpy(headbuf + 8, "WAVE", 4);
	memcpy(headbuf + 12, "fmt ", 4);
	WRITE_U32(headbuf + 16, 16);
	WRITE_U16(headbuf + 20, 1);	/* format */
	WRITE_U16(headbuf + 22, channels);
	WRITE_U32(headbuf + 24, samplerate);
	WRITE_U32(headbuf + 28, bytespersec);
	WRITE_U16(headbuf + 32, align);
	WRITE_U16(headbuf + 34, BITS);
	memcpy(headbuf + 36, "data", 4);
	WRITE_U32(headbuf + 40, size - 44);
	return write(STDOUT_FILENO, headbuf, 44) != 44 ? -1 : 1;
}
static int read_chunk(void)
{
	size_t size = 0, off = 0;
	static int eof, prev_off;

	if (stream.next_frame) { /* we already have some data */
		off = stream.bufend - stream.next_frame;
		memmove(inbuf, stream.next_frame, off);
	}
	if (eof && off == prev_off)
		return 0;
	prev_off = off;
read:
	size = read(STDIN_FILENO, inbuf + off, INBUF_SIZE - off);
//	fprintf(stderr, "size = %d, off= %i\n", size, off);
	if (size < 0)
		return size;
	if (!size) {
		eof++;
		if (!off)
			return 0;
		size = off;
		off = 0;
		goto out;
	}
	if (off + size < INBUF_SIZE * 4 / 5) {
		off += size;
		goto read;
	}
out:
	/* make buffer known to libmad */
	mad_stream_buffer(&stream, inbuf, size + off);
	stream.error = 0;
	return size;
}

int main(int argc, char *argv[])
{
	int ret = 0, i, wrote_header = 0;
	unsigned long off = 0;
	mad_timer_t timer;
	struct mad_frame frame;
	struct mad_synth synth;

	mad_stream_init(&stream);
	mad_frame_init(&frame);
	mad_synth_init(&synth);
	mad_timer_reset(&timer);
next_chunk:
	ret = read_chunk();
	if (ret <= 0) /* read error */
		goto out;
	ret = -2; /* frame decode error */
	if (mad_frame_decode(&frame, &stream)) {
		/* MAD_ERROR_BUFLEN is not really an error */
		if (MAD_RECOVERABLE(stream.error) ||
				stream.error == MAD_ERROR_BUFLEN)
			goto next_chunk;
		goto out;
	}
	if (!wrote_header) {
		fprintf(stderr, "channels: %d, %d Hz\n",
			frame.header.mode == MAD_MODE_SINGLE_CHANNEL? 1 : 2,
			frame.header.samplerate);
		write_wav_header(MAD_NCHANNELS(&frame.header),
			frame.header.samplerate);
		wrote_header = 1;
	}
	mad_timer_add(&timer, frame.header.duration);
	mad_synth_frame(&synth, &frame);
	ret = -3; /* write error */
	for (i = 0; i < synth.pcm.length; i++) {
		/* output format: unsigned 16 bit little endian */
		signed short s = MAD_TO_SHORT(synth.pcm.samples[0][i]);
		outbuf[off++] = s & 0xff;
		outbuf[off++] = s >> 8;
		if (MAD_NCHANNELS(&frame.header) == 2) { /* stereo */
			s = MAD_TO_SHORT(synth.pcm.samples[1][i]);
			outbuf[off++] = s & 0xff;
			outbuf[off++] = s >> 8;
		}
		if (off != OUTBUF_SIZE) /* buffer not yet full */
			continue;
		if (write(STDOUT_FILENO, outbuf, OUTBUF_SIZE) != OUTBUF_SIZE)
			goto out;
		off = 0;
	}
	goto next_chunk;
out:
	mad_synth_finish(&synth);
	mad_frame_finish(&frame);
	mad_stream_finish(&stream);
	if (off && ret >= 0) /* write remainder */
		ret = write(STDOUT_FILENO, outbuf, off) == off? 1 : -1;
	return ret >= 0? EXIT_SUCCESS : EXIT_FAILURE;
}
